class Photo:
	__slots__ = (
		"file_id",
		"width",
		"height",
		"file_size"
	)
	def __init__(self, file_id: str, width: int, height: int, file_size: int):
		self.file_id = file_id
		self.width = width
		self.height = height
		self.file_size = file_size

	@classmethod
	def from_dict(cls, data: dict):
		return cls(
			file_id=data.get("file_id"),
			width=data.get("width"),
			height=data.get("height"),
			file_size=data.get("file_size")
		)

	def to_dict(self):
		data = {
			"file_id": self.file_id,
			"width": self.width,
			"height": self.height,
			"file_size": self.file_size
		}
		return data


"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations
import asyncio
from typing import Callable, Dict, Tuple, List
from builtins import enumerate, reversed
from bale import (Message, Update, User, Components, RemoveComponents, Chat, Price, ChatMember, HTTPClient, Updater)


__all__ = (
    "Bot"
)


class _Loop:
    __slots__ = ()
    
    
_loop = _Loop()


class Bot:
    """This object represents a Bale Bot.

        Args:
            token (str): Bot Token.
        Attributes:
            loop (:class:`asyncio.AbstractEventLoop`)
            token (str)
            http (:class:`bale.HTTPClient`)
            user (:class:`bale.User` | None)
            updater (:class:`bale.Updater`)
            events (Dict[str, Callable])
            listeners Dict[str, List[Tuple[asyncio.Future, Callable[..., bool]]]]
            _closed (bool)
        Raises:
            :class:`bale.Error`
    """
    __slots__ = (
        "loop",
        "token",
        "loop",
        "events",
        "listeners",
        "updater",
        "user",
        "http",
        "_closed"
    )

    def __init__(self, token: str):
        self.loop = _loop
        self.token = token
        self.http: HTTPClient = HTTPClient(loop=self.loop, token=token)
        self.user = None
        self.updater = Updater(self)
        self.events: Dict[str, List[Callable]] = {}
        self.listeners: Dict[str, List[Tuple[asyncio.Future, Callable[..., bool]]]] = {}
        self._closed = False

    def listen(self, event_name):
        """Register a Event"""
        def decorator(func):
            self.add_event(event_name, func)
        return decorator

    def add_event(self, event: str, function):
        """Register an Event with event name"""
        if not asyncio.iscoroutinefunction(function):
            raise TypeError(f"{function.__name__} is not a Coroutine Function")

        if not self.events.get(event):
            self.events[event] = list()

        self.events[event].append(function)

    def remove_event(self, event: str, function = None):
        """Register an Event with event name"""
        result = self.events.get(event)
        if not result:
            raise TypeError(f"{event} not in Events")

        if not function:
            del self.events[event]
            return

        if not function in result:
            raise TypeError(f"{function.__name__} not in Event Functions")

        del self.events[event][function]

    def wait_for(self, event_name: str, check = None, timeout = None):
        """Wait for an event"""
        self.loop: asyncio.AbstractEventLoop
        future = self.loop.create_future()
        event_name = event_name.lower()
        if check:
            def event_check(*args):
                return True
            check = event_check

        listeners = self.listeners.get(event_name)
        if not listeners:
            listeners = []
            self.listeners[event_name] = listeners

        listeners.append((future, check))
        return asyncio.wait_for(future, timeout = timeout)

    async def __aenter__(self):
        loop = asyncio.get_running_loop()
        self.loop = loop
        self.http.loop = loop
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        pass

    async def close(self):
        """Close http Events and bot"""
        await self.http.close()
        self._closed = True

    def is_closed(self):
        """:class:`bool`: Connection Status"""
        return self._closed

    async def run_event(self, core, event_name, *args, **kwargs):
        """Run Core Event"""
        try:
            await core(*args, **kwargs)
        except Exception as ext:
            await self.on_error(event_name, ext)

    def call_to_run_event(self, core, event_name, *args, **kwargs):
        """Call to Run Event Method"""
        task = self.run_event(core, event_name, *args, **kwargs)
        self.loop: asyncio.AbstractEventLoop
        return self.loop.create_task(task, name = event_name)

    def dispatch(self, event_name, /, *args, **kwargs):
        """Dispatching event to all listeners"""
        method = "on_" + event_name
        listeners = self.listeners.get(event_name)
        if listeners:
            removed = []
            for index, (future, check) in enumerate(listeners):
                if future.cancelled():
                    removed.append(index)
                    continue
                try:
                    result = check(args)
                except Exception as __exception:
                    future.set_exception(__exception)
                    removed.append(index)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(index)

            if len(listeners) == len(removed):
                self.listeners.pop(event_name)
            else:
                for index in reversed(removed):
                    del listeners[index]

        events_core = self.events.get(method)
        if events_core:
            for event_core in events_core:
                self.call_to_run_event(event_core, method, *args, **kwargs)

    async def on_error(self, event_name, error):
        """a Event for get errors when exceptions"""
        print("error", event_name, error)

    async def get_bot(self) -> User:
        """Get Bot.

        Returns:
            :class:`Bale.User`: Bot User information.
        Raises:
            :class:`Bale.Error`
        """
        response, payload = await self.http.get_bot()
        return User.from_dict(data=payload.get("result"), bot=self)

    async def delete_webhook(self) -> bool:
        """This service is used to remove the webhook set for the bot.

        Returns:
            bool: If done "True" if not "False"
        """
        response, payload = await self.http.delete_webhook()
        return payload.get("result", False)

    async def send_message(self, chat_id: int | str, text: str = None, components: "Components" | "RemoveComponents" =None, reply_to_message_id: str = None) -> Message | None:
        """This service is used to send text messages.
        
        Args:
            chat_id (int): Chat ID.
            text (str): Message Text. 
            components (:class:`bot.Components`|:class:`bale.RemoveComponents`): Message Components.
            reply_to_message_id (str): Reply Message ID. 
        Raises:
            :class:`bale.Error`
        Returns:
            :class:`bale.Message`: On success, the sent Message is returned.
        """
        if not isinstance(chat_id, (str, int)):
            raise TypeError(
                f"chat_id is not a str or int. this is a {chat_id.__class__} !"
            )

        if components:
            components = components.to_dict()
        if reply_to_message_id:
            reply_to_message_id = reply_to_message_id
        response, payload = await self.http.send_message(str(chat_id), text, components=components, reply_to_message_id=reply_to_message_id)
        return Message.from_dict(data=payload["result"], bot=self)

    async def send_invoice(self, chat_id: str | int, title: str, description: str, provider_token: str, prices: List[Price], photo_url: str = None, need_name: bool = False, need_phone_number: bool = False, need_email: bool = False, need_shipping_address: bool = False, is_flexible: bool = True) -> Message | None:
        """You can use this service to send money request messages.
        Args:
            chat_id (str | int): Chat ID
            title (str): Invoice Title
            description (str): Invoice Description
            provider_token (str): You can use 3 methods to receive money: 1.Card number 2. Port number and acceptor number 3. Wallet number "Bale"
            prices (List[:class:`bale.Price`]): A list of prices.
            photo_url (str): Photo URL of Invoice. Defaults to None.
            need_name (bool): Get a name from "User"?. Defaults to False.
            need_phone_number (bool): Get a Phone number from "User"?. Defaults to False.
            need_email (bool): Get a Email from "User"?. Defaults to False.
            need_shipping_address (bool): Get a Shipping Address from "User"?. Defaults to False.
            is_flexible (bool): Is the Invoice Photo Flexible to the Payment button?. Defaults to True.
        Returns:
            :class:`Bale.Message`
        """
        if not isinstance(chat_id, (str, int)):
            raise TypeError(
                f"chat_id is not a str or int. this is a {chat_id.__class__} !"
            )
        prices = [price.to_dict() for price in prices]
        response, payload = await self.http.send_invoice(str(chat_id), title, description, provider_token, prices, photo_url, need_name, need_phone_number, need_email, need_shipping_address, is_flexible)
        return Message.from_dict(data=payload["result"], bot=self)

    async def edit_message(self, chat_id: int | str, message_id: str, text: str, components: "Components" | "RemoveComponents"=None) -> Message:
        """You can use this service to edit text messages that you have already sent through the arm.

        Args:
            chat_id (int | str): Chat Id.
            message_id (str): Message Id.
            text (str): New Content For Message.
            components (:class:`bale.Components`|:class:`bale.RemoveComponents`): Components. Defaults to None.
        Raises:
            :class:`bale.Error`
        Return:
            :class:`dict`
        """
        if not isinstance(chat_id, (str, int)):
            raise TypeError(
                f"chat_id is not a str or int. this is a {chat_id.__class__} !"
            )

        if components:
            components = components.to_dict()

        response, payload = await self.http.edit_message(str(chat_id), message_id, text, components)
        return payload

    async def delete_message(self, chat_id: str | int, message_id: str) -> bool:
        """You can use this service to delete a message that you have already sent through the arm.
        
        In Channel or Group:
            If it is a group or channel Manager, it can delete a message from (group or channel).

        In private message (PV):
            If the message was sent by a bot, it can be deleted with this method

        Args:
            chat_id (str | int): Chat ID.
            message_id (str): Message ID
        Return Type:
            bool: if done "True" if not "False"
        """
        if not isinstance(chat_id, (str, int)):
            raise TypeError(
                f"chat_id is not a str or int. this is a {chat_id.__class__} !"
            )

        response, payload = await self.http.delete_message(str(chat_id), message_id)
        return payload["result"]

    async def get_chat(self, chat_id: str) -> Chat:
        """This service can be used to receive personal information that has previously interacted with the arm.

        Args:
            chat_id (str): Chat Id.
        Raises:
            :class:`bale.Error`
        Return:
            :class:`bale.Chat`: On success, the sent Message is returned.
        """
        if not isinstance(chat_id, str):
            raise TypeError(
                f"chat_id is not a str. this is a {chat_id.__class__} !"
            )

        response, payload = await self.http.get_chat(chat_id)
        return Chat.from_dict(payload["result"], bot=self)

    async def get_chat_member(self, chat_id: str, user_id: str) -> "ChatMember":
        """
            Args:
                chat_id (str): Group ID
                user_id (str): Member ID
            Returns:
                :class:`bale.ChatMember`
            Raises:
                :class:`bale.Error`
        """
        if not isinstance(chat_id, str):
            raise TypeError(
                f"chat_id is not a str. this is a {chat_id.__class__} !"
            )

        if not isinstance(user_id, str):
            raise TypeError(
                f"user_id is not a str. this is a {user_id.__class__} !"
            )

        response, payload = await self.http.get_chat_member(chat_id=chat_id, member_id=user_id)
        return ChatMember.from_dict(payload.get("result"))

    async def get_chat_members_count(self, chat_id: str) -> int:
        """
            Args:
                chat_id (str): Group ID
            Returns:
                :int: Member Chat count
            Raises:
                :class:`bale.Error`
        """
        if not isinstance(chat_id, str):
            raise TypeError(
                f"chat_id is not a str. this is a {chat_id.__class__} !"
            )

        response, payload = await self.http.get_chat_members_count(chat_id)
        return payload["result"]

    async def get_chat_administrators(self, chat_id: str) -> list["ChatMember"]:
        """This service can be used to display admins of a group or channel.

        Args:
            chat_id (str): Group ID
        Raises:
            :class:`bale.Error`
        Returns:
            List[:class:`bale.ChatMember`]
        """
        if not isinstance(chat_id, str):
            raise TypeError(
                f"chat_id is not a str. this is a {chat_id.__class__} !"
            )

        response, payload = await self.http.get_chat_administrators(chat_id)
        return [ChatMember.from_dict(data=member_payload) for member_payload in payload["result"]]

    async def get_updates(self, offset: int = None, limit: int = None) -> list["Update"] | None:
        """Use this method to receive incoming updates using long polling.

        Args:
            offset (int, optional): Defaults to None.
            limit (int, optional): Defaults to None.
        Raises:
            :class:`bale.Error`

        Returns:
            List[:class:`bale.Update`]
        """
        if offset and not isinstance(offset, int):
            raise TypeError(
                f"offset is not a int"
            )

        if limit and not isinstance(limit, int):
            raise TypeError(
                f"limit is not a int"
            )

        response, payload = await self.http.get_updates(offset, limit)
        return [Update.from_dict(data=update_payload, bot=self) for update_payload in payload.get("result", []) if not offset or (offset and update_payload.get("update_id") > offset)]

    async def connect(self):
        """Get bot and Start Updater"""
        await self.http.start()
        self.user = await self.get_bot()
        await self.updater.start()

    def run(self):
        """
            run bot and https
        """
        async def main():
            """a runner for `asyncio.run`"""
            async with self:
                await self.connect()

        asyncio.run(main())



from __future__ import annotations


class Price:
    """This object shows a Price

        Args:
            label (str): Label Price. Defaults to None.
            amount (int): Amount Price. Defaults to None.
    """
    __slots__ = ("label", "amount")

    def __init__(self, label: str = None, amount: int = None):
        self.label = label
        self.amount = amount
        
    @classmethod
    def from_dict(cls, data):
        """
        Args:
            data (dict): Data
        """
        return cls(label=data["label"], amount=data["amount"])
    
    def to_dict(self):
        """Convert Class to dict
            Returns:
                :dict:
        """
        data: dict[str, int | str] = {
            "label": self.label,
            "amount": self.amount
        }
        return data



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bale import Bot, User


class ContactMessage:
    """This object shows a Message Contact.

        Args:
            phone_number (int)
            first_name (str): Defaults to None.
            last_name (str): Defaults to None.
            bot (:class:`bale.Bot`): Bot Object. Defaults to None.
    """
    __slots__ = (
        "phone_number",
        "first_name",
        "last_name",
        "id",
        "bot"
    )

    def __init__(self, phone_number: int, first_name: str = None, last_name: str = None, bot: "Bot" = None):
        self.phone_number = phone_number
        self.first_name = first_name
        self.last_name = last_name
        self.bot = bot

    @property
    def user(self):
        if self.id:
            return User(bot=self.bot, user_id=self.id, first_name=self.first_name, last_name=self.last_name)
        return None

    @classmethod
    def from_dict(cls, data: dict, bot: "Bot" = None):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        return cls(first_name=data["first_name"], last_name=data["last_name"],
                   bot=bot, phone_number=data["phone_number"])

    def to_dict(self):
        data = {"phone_number": self.phone_number, "first_name": self.first_name, "last_name": self.last_name}

        return data



class Audio:
    __slots__ = (
        "file_id",
        "duration",
        "title",
        "file_size",
        "mime_type",
        "bot"
    )

    def __init__(self, file_id: id, duration: int, file_size: int = None, mime_type: str = None, title: str = None):
        """This object shows a waist

        Args:
            file_id (id): Audio ID
            duration (int): Audio duration
            file_size (int): Audio Size. Defaults to None.
            mime_type (str): Audio Mime type. Defaults to None.
            title (str): Audio Title. Defaults to None.
        """
        self.file_id = file_id
        self.duration = duration

        self.title = title
        self.file_size = file_size
        self.mime_type = mime_type

    @classmethod
    def from_dict(cls, data):
        """
        Args:
            data (dict): Data
        """
        return cls(file_id=data["file_id"], duration=data["duration"], file_size=data["file_size"], title=data["title"],
                   mime_type=data["mime_type"])

    def to_dict(self):
        data = {"file_id": self.file_id if self.file_id is not None else None,
                "duration": self.duration if self.duration is not None else None,
                "title": self.title if self.title is not None else None,
                "file_size": self.file_size if self.file_size is not None else None,
                "mime_type": self.mime_type if self.mime_type is not None else None}

        return data

    def get_file(self):
        # Coming Soon...
        pass



from __future__ import annotations
from typing import Dict

__all__ = (
    "Components",
    "Keyboard",
    "InlineKeyboard",
    "RemoveComponents"
)


class Components:
    """
        Args:
            keyboards (:class:`bale.Keyboards`): Defaults to None.
            inline_keyboards (:class:`bale.InlineKeyboards`): Defaults to None.
    """

    __slots__ = (
        "keyboards",
        "inline_keyboards"
    )

    def __init__(self, keyboards=None, inline_keyboards=None):

        if not (isinstance(keyboards, list) or isinstance(inline_keyboards, list)):
            raise TypeError("The type of parameter entered is incorrect.")

        self.keyboards = []
        self.inline_keyboards = []

        if keyboards:
            for obj in keyboards:
                if not isinstance(obj, list):
                    raise TypeError("object is not a list its a {}".format(obj.__class__))

                keyboard = [i.to_dict() for i in obj]
                self.keyboards.append(keyboard)

        if inline_keyboards:
            for obj in inline_keyboards:
                if not isinstance(obj, list):
                    raise TypeError("object is not a list its a {}".format(obj.__class__))

                inline_keyboard = [i.to_dict() for i in obj]
                self.inline_keyboards.append(inline_keyboard)

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        return cls(keyboards=data["keyboard"], inline_keyboards=data["inline_keyboard"])

    def to_dict(self) -> Dict:
        """Convert Class to dict
            Returns:
                :dict:
        """
        data = {}
        keyboards = self.keyboards
        inline_keyboards = self.inline_keyboards

        if bool(keyboards):
            data["keyboard"] = keyboards

        if bool(inline_keyboards):
            data["inline_keyboard"] = inline_keyboards

        return data


class InlineKeyboard:
    """This object shows an in -line keyboard (within the message).

        Args:
            text (str):	Label text on the button.
            callback_data (str): If set, pressing the button will prompt the user to select one of their chats, open that chat and insert the bot's username and the specified inline query in the input field. Can be empty, in which case just the bot's username will be inserted. Defaults to None.
            url (str): HTTP url to be opened when the button is pressed. Defaults to None.
            switch_inline_query (str): If set, pressing the button will prompt the user to select one of their chats, open that chat and insert the bot's username and the specified inline query in the input field. Can be empty, in which case just the bot's username will be inserted. Defaults to None.
            switch_inline_query_current_chat (str): If set, pressing the button will insert the bot's username and the specified inline query in the current chat's input field. Can be empty, in which case only the bot's username will be inserted. Defaults to None.
    """
    __slots__ = (
        "text", "callback_data", "url", "switch_inline_query", "switch_inline_query_current_chat"
    )

    def __init__(self, text: str, callback_data: str = None, url: str = None, switch_inline_query: str = None,
                 switch_inline_query_current_chat: str = None):
        self.text = text
        self.callback_data = callback_data if callback_data is not None else None
        self.url = url if url is not None else None
        self.switch_inline_query = switch_inline_query if switch_inline_query is not None else switch_inline_query
        self.switch_inline_query_current_chat = switch_inline_query_current_chat if switch_inline_query_current_chat is not None else None

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        if not data.get("text") or not data.get("callback_data"):
            return None
        return cls(text=data["text"], callback_data=data.get("callback_data"), url=data.get("url"),
                   switch_inline_query=data.get("switch_inline_query"),
                   switch_inline_query_current_chat=data.get("switch_inline_query_current_chat"))

    def to_dict(self) -> dict:
        """Convert Class to dict
            Returns:
                :dict:
        """
        data = {
            "text": self.text
        }

        if self.callback_data:
            data["callback_data"] = self.callback_data

        if self.url:
            data["url"] = self.url

        if self.switch_inline_query:
            data["switch_inline_query"] = self.switch_inline_query

        if self.switch_inline_query_current_chat:
            data["switch_inline_query_current_chat"] = self.switch_inline_query_current_chat

        return data


class Keyboard:
    """This object shows a keyboard

        Args:
            text (str): Keyboard Text.
    """
    __slots__ = (
        "text"
    )

    def __init__(self, text: str):
        self.text = text

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        if not data.get("text"):
            return None
        return cls(text=data["text"])

    def to_dict(self):
        """Convert Class to dict
        Returns:
            :dict:
        """
        data = {
            "text": self.text
        }
        return data


class RemoveComponents:
    """This object shows a remove keyboard"""

    def to_dict(self) -> dict:
        """Convert Class to dict
        Returns:
            :dict:
        """
        return {"inline_keyboard": None, "keyboard": None}



import sys
from .version import __version__


def main():
    print("Python-Bale-Bot By Kian Ahmadian")
    print("Python Version: ", sys.version)
    print("Package Version: ", __version__)


if __name__ == '__main__':
    main()



from .document import Document
from .audio import Audio
from .contact import ContactMessage
from .location import Location
from .photo import Photo

__all__ = (
	"Document",
	"Audio",
	"ContactMessage",
	"Location",
	"Photo"
)



from bale import (AdminPermissions, User)


class Role:
    """Member Role"""
    __slots__ = ()
    OWNER = "creator"
    ADMIN = "administrator"


class ChatMember:
    """This object shows a user in chat

        Args:
            role (str): User Role. Defaults to None.
            user (:class:`bale.User`): User. Defaults to None.
            permissions (:class:`bale.AdminPermissions`): User Permissions. Defaults to None.
    """
    __slots__ = (
        "role", "_user", "permissions"
    )

    def __init__(self, role: str = None, user=None, permissions=None):
        self.role = role
        self._user = user
        self.permissions = permissions

    def is_admin(self):
        """if the member was the admin, it will be returned "True" and otherwise "False".

        Returns:
            bool: if the member was the admin, it will be returned "True" and otherwise "False".
        """
        return self.role == Role.ADMIN or self.role == Role.OWNER

    def is_owner(self):
        """if the member was the owner, it will be returned "True" and otherwise "False".

        Returns:
            bool: if the member was the owner, it will be returned "True" and otherwise "False".
        """
        return self.role == Role.OWNER

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        return cls(permissions=AdminPermissions.from_dict(data), user=User.from_dict(data.get("user")), role=data.get("status"))



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from bale import Bot

from bale import (Chat, ChatType, User, Document, ContactMessage, Photo, Invoice)


class Message:
    """This object shows a message

    Args:
        message_id (str): Message ID.
        date (datetime.datetime): When the message has been sent.
        text (str): Message Text. Defaults to None.
        caption (str): Message caption. Defaults to None.
        from_user (:class:`bale.User`): The user who has sent the message. Defaults to None.
        forward_from (:class:`bale.User`) no description. Defaults to None.
        contact (:class:`bale.ContactMessage`): Defaults to None.
        chat (:class:`bale.Chat`): The chat where the message is sent. Defaults to None.
        reply_to_message (:class:`bale.Message`)
        new_chat_members (:class:`bale.User`): User (An) who entered the chat. Defaults to None.
        left_chat_member (:class:`bale.User`): A user out of chat. Defaults to None.
        invoice (:class:`bale.Invoice`): Message invoice.
        bot (:class:`bale.Bot`): Bot object. Defaults to None.
    """
    __slots__ = (
        "text", "caption", "from_user", "_author", "contact", "chat", "message_id", "forward_from", "forward_from_message_id", "date_code", "date", "edit_date",
        "audio", "document", "photos", "voice", "location", "invoice", "new_chat_members", "left_chat_member",
        "reply_to_message", "invoice", "bot"
    )

    def __init__(self, message_id: str, date: datetime, text: str = None, caption: str = None,
                 forward_from: "User" = None, forward_from_message_id: str = None, from_user: "User" = None, document: "Document" = None, contact: "ContactMessage" = None, chat: "Chat" = None,
                 photos: List["Photo"] = None, reply_to_message: "Message" = None, invoice: "Invoice" = None, bot: 'Bot' = None, **options):
        self.message_id: str = message_id if message_id is not None else None
        self.date = date if date is not None else None

        self.text: str | None = text if text is not None else None
        self.chat: Chat | None = chat if chat is not None else None
        self.reply_to_message: Message | None = reply_to_message if reply_to_message is not None else reply_to_message
        self.from_user: User | None = from_user if from_user is not None else None
        self.forward_from: User | None = forward_from if forward_from is not None else None
        self.forward_from_message_id: str = forward_from_message_id if forward_from_message_id is not None else None
        self.caption: str | None = caption if caption is not None else None
        self.document = document if document is not None else None
        self.photos = photos if photos is not None else None
        self.contact: ContactMessage | None = contact if contact is not None else None
        self.new_chat_members: List[User] | None = options.get("new_chat_members")
        self.left_chat_member: User | None = options.get("left_chat_member")
        self.invoice = invoice
        self.bot: Bot = bot if bot is not None else None

    @property
    def author(self):
        return self.from_user

    @property
    def content(self):
        return self.caption or self.text

    @content.setter
    def content(self, _value):
        if not isinstance(_value, str):
            raise TypeError(f"{_value} is not str. this is a {_value.__class__ if _value else _value}")
        if self.caption:
            self.caption = _value
        elif self.text:
            self.text = _value

    @property
    def chat_id(self):
        if self.chat is not None:
            return self.chat.chat_id
        return None

    @property
    def reply_to_message_id(self):
        if self.reply_to_message:
            return self.reply_to_message.message_id
        return None

    @reply_to_message_id.setter
    def reply_to_message_id(self, _value):
        if not isinstance(_value, int):
            raise TypeError(f"{_value} is not str. this is a {_value.__class__ if _value else _value}")

        if self.reply_to_message:
            self.reply_to_message.message_id = _value

    @classmethod
    def from_dict(cls, data: dict, bot):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        options = {}
        if data.get("new_chat_members"):
            options["new_chat_members"] = [User.from_dict(bot=bot, data=i) for i in data.get("new_chat_members")]
        if data.get("left_chat_member"):
            options["left_chat_member"] = User.from_dict(bot=bot, data=data.get("left_chat_member"))

        return cls(bot=bot, message_id=str(data.get("message_id")),
                   chat=Chat.from_dict(bot=bot, data=data.get("chat")) if data.get("chat") else None,
                   reply_to_message=Message.from_dict(bot=bot, data=data.get("reply_to_message")) if data.get(
                       "reply_to_message") else None, date=datetime.fromtimestamp(int(data.get("date"))), text=data.get("text"),
                   from_user=User.from_dict(bot=bot, data=data.get("from")) if data.get("from") else None,
                   forward_from=User.from_dict(bot=bot, data=data.get("forward_from")) if data.get("forward_from") else None,
                   forward_from_message_id=str(data.get("forward_from_message_id")) if data.get("forward_from_message_id") else None,
                   document=Document.from_dict(data=data.get("document")) if data.get("document") else None,
                   photo=[Photo.from_dict(data=photo_payload) for photo_payload in data.get("photo")] if data.get("photo") else None,
                   invoice=Invoice.from_dict(data=data.get("invoice")) if data.get("invoice") else None, **options)

    def to_dict(self):
        data = {"message_id": self.message_id, "date": self.date, "text": self.text}

        if self.chat:
            data["chat"] = self.chat.to_dict()
        if self.from_user:
            data["from"] = self.from_user.to_dict()
        if self.caption:
            data["caption"] = self.caption
        if self.document:
            data["document"] = self.document.to_dict()
        if self.photos:
            data["photo"] = [photo.to_dict() for photo in self.photos]
        if self.contact:
            data["contact"] = self.contact.to_dict()
        if self.new_chat_members:
            data["new_chat_members"] = self.new_chat_members
        if self.forward_from:
            data["forward_from"] = self.forward_from.to_dict()
        if self.left_chat_member:
            data["left_chat_member"] = self.left_chat_member.to_dict()
        if self.reply_to_message_id:
            data["reply_to_message_id"] = self.reply_to_message_id

        return data

    async def reply(self, text: str, components=None):
        """:meth:`bale.Bot.send_message`

        Args:
            text (str): Message Text.
            components (:class:`bot.Components`|:class:`bale.RemoveComponents`): Message Components.
        Raises:
            :class:`bale.Error`
        Returns:
            :class:`bale.Message`: On success, the sent Message is returned."""
        result = await self.bot.send_message(chat_id=self.chat.chat_id, text=text, components=components,
                                       reply_to_message_id=str(self.message_id) if not self.chat.type.is_group_chat() else None)
        return result

    async def reply_invoice(self, title: str, description: str, provider_token: str, prices, photo_url=None, need_name=False,
                      need_phone_number=False, need_email=False, need_shipping_address=False, is_flexible=True):
        """:meth:`bale.Bot.send_invoice`

        Args:
            title (str): Invoice Title
            description (str): Invoice Description
            provider_token (str): You can use 3 methods to receive money: 1.Card number 2. Port number and acceptor number 3. Wallet number "Bale"
            prices (List[:class:`bale.Price`]): A list of prices.
            photo_url (str): Photo URL of Invoice. Defaults to None.
            need_name (bool): Get a name from "User"?. Defaults to False.
            need_phone_number (bool): Get a Phone number from "User"?. Defaults to False.
            need_email (bool): Get a Email from "User"?. Defaults to False.
            need_shipping_address (bool): Get a Shipping Address from "User"?. Defaults to False.
            is_flexible (bool): Is the Invoice Photo Flexible to the Payment button?. Defaults to True.
        Returns:
            :class:`Bale.Message`
        """
        message = await self.bot.send_invoice(chat_id=self.chat.chat_id, title=title, description=description,
                                        provider_token=provider_token, prices=prices, photo_url=photo_url,
                                        need_name=need_name, need_email=need_email, need_phone_number=need_phone_number,
                                        need_shipping_address=need_shipping_address, is_flexible=is_flexible)
        return message

    async def edit(self, text: str = None, components=None):
        """:meth:`bale.Bot.edit_message`

        Args:
            text (str): New Content For Message.
            components (:class:`bale.Components`|:class:`bale.RemoveComponents`): Components.
        Raises:
            :class:`bale.Error`
        Return:
            :class:`dict`
        """
        result = await self.bot.edit_message(self.chat.chat_id, self.message_id, text, components)
        self.text = text
        return result

    async def delete(self):
        """:meth:`bale.Bot.delete_message`

        Return:
            bool: if done "True" if not "False"
        """
        message = await self.bot.delete_message(chat_id=self.chat.chat_id, message_id=str(self.message_id))
        return message

    def __str__(self):
        return str(self.message_id)



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from bale.version import BALE_API_BASE_URL
import aiohttp
from ..error import (NetworkError, HTTPException, TimeOut, NotFound, Forbidden, APIError, BaleError)


__all__ = ("HTTPClient",)


class Route:
	"""Route Class for http"""
	BASE = BALE_API_BASE_URL

	__slots__ = (
		"method",
		"path",
		"url",
		"token"
	)

	def __init__(self, method: str, path: str, token: str):
		if not isinstance(token, str):
			raise TypeError("token is not str!\ntoken is a {}".format(token.__class__))
		self.url = self.BASE + "bot" + token + "/" + path
		self.method = method
		self.path = path
		self.token = token


class HTTPClient:
	"""Send a Request to BALE API Server"""

	__slots__ = (
		"loop",
		"token",
		"__session"
	)

	def __init__(self, loop, token=None):
		self.__session = None
		self.loop = loop
		self.token = token

	def reload_session(self):
		"""Reset Session"""
		if self.__session and self.__session.closed:
			self.__session = aiohttp.ClientSession(loop=self.loop, connector=aiohttp.TCPConnector(limit=0))

	async def start(self):
		"""Start Http client"""
		self.__session = aiohttp.ClientSession(loop=self.loop, connector=aiohttp.TCPConnector(limit=0))

	async def close(self):
		"""Close Session connection"""
		if self.__session:
			await self.__session.close()

	async def request(self, route: Route, **kwargs):
		url = route.url
		method = route.method

		try:
			async with self.__session.request(method=method, url=url, **kwargs) as response:
				response: aiohttp.ClientResponse = response
				if response.status == 200:
					payload = await response.json()
					if not payload.get("ok"):
						raise APIError(
							str(payload.get("error_code")) + payload.get("description")
						)
					return response, payload
				elif response.status == 404:
					raise NotFound()
				elif response.status == 403:
					raise Forbidden()
				else:
					raise HTTPException(response, response.json())
		except aiohttp.client_exceptions.ClientConnectorError as error:
			raise NetworkError(str(error))
		except aiohttp.client_exceptions.ServerTimeoutError:
			raise TimeOut()
		except aiohttp.client_exceptions.ClientOSError as error:
			raise BaleError(error)

	def send_message(self, chat_id, text, *, components=None, reply_to_message_id=None):
		payload = {
			"chat_id": chat_id,
			"text": text
		}
		if components:
			payload["reply_markup"] = components
		if reply_to_message_id:
			payload["reply_to_message_id"] = reply_to_message_id

		return self.request(Route("POST", "sendMessage", self.token), json=payload)

	def send_invoice(self, chat_id, title, description, provider_token, prices, photo_url=None, need_name=False, need_phone_number=False, need_email=False, need_shipping_address=False, is_flexible=True):
		payload = {"chat_id": chat_id, "title": title, "description": description, "provider_token": provider_token, "prices": prices}
		if photo_url:
			payload["photo_url"] = photo_url
		payload["need_name"] = need_name
		payload["need_phone_number"] = need_phone_number
		payload["need_email"] = need_email
		payload["need_shipping_address"] = need_shipping_address
		payload["is_flexible"] = is_flexible

		return self.request(Route("POST", "sendInvoice", self.token), json=payload)

	def edit_message(self, chat_id, message_id, text, components=None):
		payload = {
			"chat_id": chat_id,
			"message_id": message_id,
			"text": text
		}
		if components:
			payload["reply_markup"] = components

		return self.request(Route("POST", "editMessageText", self.token), json=payload)

	def delete_message(self, chat_id, message_id):
		payload = {
			"chat_id": chat_id,
			"message_id": message_id
		}
		return self.request(Route("GET", "deletemessage", self.token), params=payload)

	def get_updates(self, offset=None, limit=None):
		payload = {}
		if offset:
			payload["offset"] = offset
		if limit:
			payload["limit"] = limit
		return self.request(Route("POST", "getupdates", self.token), data=payload)

	def delete_webhook(self):
		return self.request(Route("GET", "deleteWebhook", self.token))

	def get_bot(self):
		return self.request(Route("GET", "getme", self.token))

	def get_chat(self, chat_id):
		return self.request(Route("GET", "getchat", self.token), params=dict(chat_id=chat_id))

	def get_chat_administrators(self, chat_id):
		return self.request(Route("GET", "getChatAdministrators", self.token), params=dict(chat_id=chat_id))

	def get_chat_members_count(self, chat_id):
		return self.request(Route("GET", "getChatAdministrators", self.token), params=dict(chat_id=chat_id))

	def get_chat_member(self, chat_id, member_id):
		return self.request(Route("GET", "getChatMember", self.token), params=dict(chat_id=chat_id, user_id=member_id))



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bale import Bot, Message


__all__ = (
    "Chat",
    "ChatType"
)

class ChatType:
    """This object indicates a Chat Type.

        Args:
            _type (str): The type of Chat
    """
    PRIVATE = "private"
    GROUP = "group"
    __slots__ = (
        "_type",
    )

    def __init__(self, _type: str):
        self._type = _type

    def is_private_chat(self):
        """Return True if Chat Type is Private"""
        return self._type == self.PRIVATE

    def is_group_chat(self):
        """Return True if Chat Type is Group"""
        return self._type == self.GROUP

    def __eq__(self, other):
        return self._type == other

    def __ne__(self, other):
        return not self.__eq__(other)

class Chat:
    """This object indicates a chat.

        Args:
            chat_id (str): Chat ID.
            _type (:class:`bale.ChatType`): Chat Type.
            title (str): Chat Title.
            username (str): Chat Username (for DM or PV).
            first_name (str): First name Chat (for DM or PV).
            last_name (str): Last name Chat (for DM or PV).
            pinned_message (:class:`bale.Message`): Pinned messages in chat. Defaults to None.
            all_members_are_administrators (bool): Does everyone have admin access?. Defaults to True. (for Group)
            bot (:class:`bale.Bot`): Bot Object. Defaults to None.
    """
    __slots__ = (
        "chat_id",
        "_type",
        "title",
        "username",
        "first_name",
        "last_name",
        "pinned_message",
        "all_members_are_administrators",
        "bot"
    )

    def __init__(self, chat_id: int | str, _type: "ChatType", title: str, username: str, first_name: str, last_name: str,
                 pinned_message: Message | None = None, all_members_are_administrators: bool = True, bot: 'Bot' = None):
        self.chat_id = chat_id
        self._type = _type
        self.title = title
        self.username = username
        self.first_name = first_name
        self.last_name = last_name
        self.pinned_message = pinned_message
        self.all_members_are_administrators = all_members_are_administrators
        self.bot = bot

    @property
    def type(self):
        """Get chat type"""
        return self._type

    async def send(self, text: str, components=None):
        """Send Message in Chat
        
        Args:
            text (str): Message Text.
            components (:class:`bale.Components`|:class:`bale.RemoveComponents`): Defaults to None.
        Returns:
            :class:`bale.Message`
        """
        return await self.bot.send_message(self.chat_id, text, components=components)

    async def get_chat_member(self, user_id: str):
        """:meth:`bale.Bot.get_chat_member`.

            Args:
                user_id (str): User ID.
            Returns:
                :class:`bale.ChatMember`
        """
        return await self.bot.get_chat_member(chat_id=self.chat_id, user_id=user_id)

    async def get_chat_members_count(self):
        """:meth:`bale.Bot.get_chat_members_count`.

            Returns:
                :int:
        """
        return await self.bot.get_chat_members_count(chat_id=self.chat_id)

    async def get_chat_administrators(self):
        """:meth:`bale.Bot.get_chat_administrators`.

        Raises:
            :class:`bale.Error`
        Returns:
            List[:class:`bale.ChatMember`]
        """
        return await self.bot.get_chat_administrators(self.chat_id)

    @classmethod
    def from_dict(cls, data: dict, bot):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        return cls(bot=bot, chat_id=data.get("id"), _type=ChatType(data.get("type")), title=data.get("title"),
                   username=data.get("username"), first_name=data.get("first_name"), last_name=data.get("last_name"),
                   pinned_message=Message.from_dict(bot=bot, data=data.get("pinned_message")) if data.get("pinned_message") else None,
                   all_members_are_administrators=data.get("all_members_are_administrators", True))

    def to_dict(self):
        """Convert Class to dict
        Returns:
            :dict:
        """
        data = {
            "id": self.chat_id,
            "type": self.type,
            "title": self.title,
            "username": self.username,
            "first_name": self.first_name,
            "last_name": self.last_name
        }
        if self.pinned_message:
            data["pinned_message"] = self.pinned_message.to_dict()

        return data

    def __str__(self):
        return (str(self.username) + "#" + str(self.chat_id) if self.username else str(self.first_name) + " " + str(
            self.last_name))

    def __eq__(self, other):
        return isinstance(other, Chat) and self.chat_id == other.chat_id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return f"<Chat first_name={self.first_name} last_name={self.last_name} user_id={self.chat_id} username={self.username} title={self.title}>"



from .version import __version__
from .request.http import HTTPClient
from .components import Components, RemoveComponents, InlineKeyboard, Keyboard
from .attachments.audio import Audio
from .attachments.location import Location
from .attachments.document import Document
from .attachments.photo import Photo
from .payments import Price, Invoice
from .user import User
from .attachments.contact import ContactMessage
from .chat import Chat, ChatType
from .message import Message
from .permissions import AdminPermissions
from .chatmember import ChatMember, Role
from .callbackquery import CallbackQuery
from .update import Update
from .updater import Updater
from .error import BaleError, APIError
from .bot import Bot


__all__ = (
    "Bot",
    "CallbackQuery",
    "Update",
    "Chat",
    "ChatType",
    "Message",
    "User",
    "Updater",
    "Components",
    "RemoveComponents",
    "Keyboard",
    "InlineKeyboard",
    "Location",
    "Audio",
    "Document",
    "Photo",
    "ContactMessage",
    "Invoice",
    "Price",
    "HTTPClient",
    "BaleError",
    "APIError",
    "AdminPermissions",
    "ChatMember", 
    "Role"
)

__title__ = "python-bale-bot"
__author__ = "Kian Ahmadian"



"""MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""
import asyncio
from typing import TYPE_CHECKING
if TYPE_CHECKING:
	from bale import Bot

__all__ = (
	"Updater",
)

class Updater:
	"""This object represents a Bale Bot.

		Attributes:
			bot (:class:`bale.Bot`): The bot used with this Updater.
            update_queue (:class:`asyncio.Queue`): Queue for the updates.
            _last_offset (int | None): Last of Offset for get updates.
            _is_running (bool): get status of updater.
        Args:
            bot (:class:`bale.Bot`): Bot.
        Raises:
            :class:`bale.Error`
    """
	__slots__ = (
		"bot",
		"update_queue",
		"_last_offset",
		"_is_running"
	)

	def __init__(self, bot: "Bot"):
		self.bot = bot
		self.update_queue = asyncio.Queue()
		self._last_offset = None
		self._is_running = False

	async def start(self):
		"""Start poll event function"""
		if self._is_running:
			raise RuntimeError("Updater is running")
		self._is_running = True
		self.bot.dispatch("on_ready")
		await self.poll_event()

	async def poll_event(self):
		"""A loop for get updates in dispatch"""
		if not self._is_running:
			raise RuntimeError("Updater is running")
		while self._is_running:
			if self.bot.is_closed():
				raise RuntimeError("Bot is Closed!")

			if not self._is_running:
				raise RuntimeError("Updater is not running")

			updates = await self.bot.get_updates(offset=self._last_offset)

			for update in updates:
				self.bot.dispatch("update", update)
				if update.type == "callback_query":
					self.bot.dispatch("callback", update, update.callback_query)
				elif update.type == "message":
					self.bot.dispatch("message", update, update.message)

			self._last_offset = updates[-1].update_id if bool(updates) else self._last_offset

	def stop(self):
		"""Stop running and Stop `poll_event` loop"""
		if not self._is_running:
			raise RuntimeError("Updater is not running")

		self._is_running = False



from __future__ import annotations


class Invoice:
	"""This object shows Invoice

        Args:
            title (str): Invoice title.
            description (str): Invoice Description.
            start_parameter (str)
            currency (str)
            total_amount (int)
    """
	__slots__ = (
		"title",
		"description",
		"start_parameter",
		"currency",
		"total_amount"
	)
	def __init__(self, title: str, description: str, start_parameter: str, currency: str, total_amount: int):
		self.title = title
		self.description = description
		self.start_parameter = start_parameter
		self.currency = currency
		self.total_amount = total_amount

	@classmethod
	def from_dict(cls, data: dict):
		return cls(
			title=data.get("title"),
			description=data.get("description"),
			start_parameter=data.get("start_parameter"),
			currency=data.get("currency"),
			total_amount=data.get("total_amount")
		)



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Dict
from bale import (Message, CallbackQuery)
if TYPE_CHECKING:
    from bale import Bot


class UpdateType:
    """Update Type"""
    MESSAGE = "message"
    CALLBACK = "callback_query"
    UNKNOWN = "unknown"


class Update:
    """This object shows an update.

    Args:
        update_id (int): Update ID
        callback_query (:class:`bale.CallbackQuery`): Defaults to None.
        message (:class:`bale.Message`): Defaults to None.
        edited_message (:class:`bale.Message`): Defaults to None.
        bot (:class:`bale.Bot`): Defaults to None.
        raw_data (dict): Defaults to None.
    """
    __slots__ = (
        "update_id",
        "_type",
        "message",
        "callback_query",
        "edited_message",
        "bot",
        "raw_data"
    )

    def __init__(self, update_id: int, callback_query: "CallbackQuery" = None, message: "Message" = None,
                 edited_message: "Message" = None, bot: 'Bot' = None, raw_data: dict = None):
        self.update_id = int(update_id)
        self.bot = bot
        self.raw_data = raw_data
        self.callback_query = callback_query if callback_query is not None else None
        self.message = message if message is not None else None
        self.edited_message = edited_message if edited_message is not None else None

    @property
    def type(self) -> Literal['callback_query', 'message', 'unknown']:
        """Chat Type

        Returns:
            str: Literal['callback_query', 'message', 'unknown']
        """
        if self.callback_query is not None:
            return "callback_query"
        elif self.message is not None:
            return "message"
        return "unknown"

    @classmethod
    def from_dict(cls, data: dict, bot: "Bot"):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        callback_query, message, edited_message = None, None, None

        if data.get("callback_query"):
            callback_query = CallbackQuery.from_dict(data.get("callback_query"), bot=bot)
        if data.get("message"):
            message = Message.from_dict(data.get("message"), bot=bot)
        if data.get("edited_message"):
            edited_message = Message.from_dict(data=data.get("edited_message"), bot=bot)

        return cls(update_id=data["update_id"], message=message, callback_query=callback_query, edited_message=edited_message,
                   raw_data=data, bot=bot)

    def to_dict(self) -> Dict:
        data = {}

        if self.type:
            data["type"] = self.type
        if self.callback_query:
            data["callback_query"] = self.callback_query.to_dict()
        if self.message:
            data["message"] = self.message.to_dict()
        if self.edited_message:
            data["edited_message"] = self.edited_message.to_dict()

        return data

    def __eq__(self, other):
        return isinstance(other, Update) and self.update_id == other.update_id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        if not isinstance(other, Update):
            raise NotImplemented

        return self.update_id <= other.update_id

    def __ge__(self, other):
        if not isinstance(other, Update):
            raise NotImplemented

        return self.update_id >= other.update_id

    def __lt__(self, other):
        if not isinstance(other, Update):
            raise NotImplemented

        return self.update_id < other.update_id

    def __gt__(self, other):
        return not self.__lt__(other)

    def __repr__(self):
        return f"<Update type={self.type} message={self.message}>"



class Document:
    """This object shows a Document.

        Args:
            file_id (str): Defaults to None.
            file_name (str): Defaults to None.
            mime_type (str): Defaults to None.
            file_size (int): Defaults to None.
    """
    __slots__ = (
        "file_id",
        "file_name",
        "mime_type",
        "file_size"
    )

    def __init__(self, file_id: str = None, file_name: str = None, mime_type: str = None, file_size: int = None):
        self.file_id = file_id if file_id is not None else None
        self.file_name = file_name if file_name is not None else None
        self.mime_type = mime_type if mime_type is not None else None
        self.file_size = file_size if file_size is not None else None

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        return cls(file_id=data.get("file_id"), file_name=data.get("file_name"),
                   mime_type=data.get("mime_type"), file_size=data.get("file_size"))

    def to_dict(self):
        data = {"file_id": self.file_id, "file_name": self.file_name, "mime_type": self.mime_type, "file_size": self.file_size}

        return data



from setuptools import setup, find_packages

def get_readme_file() -> str:
    """Get Readme File"""
    with open("README.md", encoding="utf-8") as f:
        readme_file = f.read()
    return readme_file

if __name__ == "__main__":
    setup(
        name="python-bale-bot",
        version="2.3.1",
        platforms=["Windows"],
        fullname="python-bale-bot",
        description="An API wrapper for Bale written in Python",
        author="Kian Ahmadian",
        license="MIT License",
        project_urls={
            "Source Code": "https://github.com/kianahmadian/python-bale-bot/",
            "Documentation": "https://python-bale-bot.readthedocs.io/en/latest/",
            "Bug Tracker": "https://github.com/kianahmadian/python-bale-bot/issues"
        },
        keywords=["bale", "bale-bot", "framework"],
        python_requires='>=3.8',
        include_package_data=True,
        url="https://github.com/kianahmadian/python-bale-bot/",
        packages=find_packages(),
        long_description=get_readme_file(),
        long_description_content_type='text/markdown',
        install_requires=["aiohttp<3.8.0,>=3.6.0", "asyncio~=3.4.3", "setuptools~=60.2.0"]
    )



print(r"""
____    __        ____   ____  ____  _____
|___|  /__\  |    |___   |___| |   |   |
|___| /    \ |___ |___   |___| |___|   |

""")



class BaleError(Exception):
    __slots__ = (
        "message",
    )

    def __init__(self, message):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.message}')"

    def __reduce__(self):
        return self.__class__, (self.message,)


class InvalidToken(BaleError):
    __slots__ = ("_message",)

    def __init__(self, message):
        self._message = message

        super().__init__("Invalid Token" if self._message is not None else self._message)


class APIError(BaleError):
    __slots__ = ("_message",)

    def __init__(self, message):
        self._message = message

        super().__init__(self._message)


class NetworkError(BaleError):
    __slots__ = ()


class TimeOut(BaleError):
    __slots__ = ()

    def __init__(self):
        super().__init__("Time Out")


class NotFound(BaleError):
    __slots__ = ()

    def __init__(self):
        super().__init__("Not Found")


class Forbidden(BaleError):
    __slots__ = ()

    def __init__(self):
        super(self).__init__("Forbidden")


class HTTPException(BaleError):
    __slots__ = ()

    def __init__(self, response, data):
        super().__init__(str(response) + str(data))



"""
    MIT License

    Copyright (c) 2022 kian Ahmadian

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
from bale import Components
if TYPE_CHECKING:
    from bale import Bot


class User:
    """This object shows a user.
        Args:
            user_id (int): User ID.
            first_name (str): first user name
            last_name (str): last user name. Defaults to None.
            username (str): Username of the user. Defaults to None.
            bot (:class:`bale.Bot`): Bot Object. Defaults to None.
    """
    __slots__ = (
        "first_name",
        "last_name",
        "username",
        "user_id",
        "bot"
    )

    def __init__(self, user_id: int, first_name: str, last_name: str = None, username: str = None, bot: 'Bot' = None):
        self.first_name = first_name
        self.last_name = last_name
        self.username = username
        self.user_id = user_id
        self.bot = bot

    @property
    def mention(self) -> str | None:
        """Mention User"""
        return f"@{self.username}" if self.username else None

    @property
    def link(self):
        """Get user link in bale"""
        if self.username:
            return "https://ble.ir/@{username}".format(username=self.username)
        return None

    async def send(self, text: str, components=None):
        """Send a Text Message to User

        Args:
            text (str): Message Text.
            components (:class:`bale.Components`|:class:`bale.RemoveComponents`): Message Components. Defaults to None.
        Returns:
            :class:`bale.Message`
        """
        from bale import Message
        if components:
            components = components.to_dict()
        response, payload = await self.bot.http.send_message(str(self.user_id), text, components=components)
        return Message.from_dict(data=payload["result"], bot=self.bot)

    @classmethod
    def from_dict(cls, data: dict, bot=None):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        return cls(username=data.get("username"), first_name=data.get("first_name"), last_name=data.get("last_name"),
                   user_id=data.get("id"), bot=bot)

    def to_dict(self):
        """Convert Class to dict
        Returns:
            :dict:
        """
        data = {"first_name": self.first_name if self.first_name is not None else None,
                "last_name": self.last_name if self.last_name is not None else None,
                "username": self.username if self.username is not None else None,
                "id": self.user_id if self.user_id is not None else None}

        return data

    def __str__(self):
        return (str(self.username) + "#" + str(self.user_id) if self.username else str(self.first_name) + " " + str(
            self.last_name))

    def __eq__(self, other: User):
        return isinstance(other, User) and self.user_id == other.user_id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__str__())

    def __repr__(self):
        return f"<User first_name={self.first_name} last_name={self.last_name} user_id={self.user_id} username={self.username}>"



class AdminPermissions:
    """This object shows the permissions and permissions of an admin or a member in a group (or channel).

        Args:
            can_be_edited (bool): Can you edit?. Defaults to False.
            can_change_info (bool): Can you edit group information? Defaults to False.
            can_post_messages (bool): Can he post a message?. Defaults to False.
            can_edit_messages (bool): Can you edit your message? Defaults to False.
            can_delete_messages (bool): Can it erase messages? Defaults to False.
            can_invite_users (bool): Can it invite users to chat? Defaults to False.
            can_restrict_members (bool): Defaults to False.
            can_pin_messages (bool): Can you pin your message? Defaults to False.
            can_promote_members (bool): Defaults to False.
            can_send_messages (bool): Can he send a message?. Defaults to False.
            can_send_media_messages (bool): Can it attach a file with the message? Defaults to False.
    """
    __slots__ = (
        "can_be_edited",
        "can_change_info",
        "can_post_messages",
        "can_edit_messages",
        "can_delete_messages",
        "can_invite_users",
        "can_restrict_members",
        "can_pin_messages",
        "can_promote_members",
        "can_send_messages",
        "can_send_media_messages"
    )

    def __init__(self, can_be_edited: bool = False, can_change_info: bool = False, can_post_messages: bool = False,
                 can_edit_messages: bool = False, can_delete_messages: bool = False, can_invite_users: bool = False,
                 can_restrict_members: bool = False, can_pin_messages: bool = False, can_promote_members: bool = False,
                 can_send_messages: bool = False, can_send_media_messages: bool = False):
        self.can_be_edited = can_be_edited
        self.can_change_info = can_change_info
        self.can_post_messages = can_post_messages
        self.can_edit_messages = can_edit_messages
        self.can_delete_messages = can_delete_messages
        self.can_invite_users = can_invite_users
        self.can_restrict_members = can_restrict_members
        self.can_pin_messages = can_pin_messages
        self.can_promote_members = can_promote_members
        self.can_send_messages = can_send_messages
        self.can_send_media_messages = can_send_media_messages

    @classmethod
    def from_dict(cls, data: dict):
        """
        Args:
            data (dict): Data
        """
        return cls(can_be_edited=data.get("can_be_edited"), can_change_info=data.get("can_change_info"),
                   can_post_messages=data.get("can_post_messages"), can_edit_messages=data.get("can_edit_messages"),
                   can_delete_messages=data.get("can_delete_messages"), can_invite_users=data.get("can_invite_users"),
                   can_restrict_members=data.get("can_restrict_members"), can_pin_messages=data.get("can_pin_messages"),
                   can_promote_members=data.get("can_promote_members"), can_send_messages=data.get("can_send_messages"),
                   can_send_media_messages=data.get("can_send_media_messages"))



class Location:
    __slots__ = (
        "longitude",
        "latitude"
    )

    def __init__(self, longitude: int, latitude: int):
        """This object shows an end

        Args:
            longitude (int): Location longitude
            latitude (int): Location latitude
        """
        self.longitude = longitude
        self.latitude = latitude

    @property
    def link(self):
        return f"https://maps.google.com/maps?q=loc:{self.longitude},{self.latitude}"

    @classmethod
    def from_dict(cls, data):
        """
        Args:
            data (dict): Data
        """
        return cls(longitude=data["longitude"], latitude=data["latitude"])

    def to_dict(self):
        data = {"longitude": self.longitude if self.longitude is not None else None,
                "latitude": self.latitude if self.latitude is not None else None}
        return data



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from bale import Bot
from bale import User, Message

__all__ = (
    "CallbackQuery"
)


class CallbackQuery:
    """This object represents an incoming callback query from a callback button in an inline keyboard.

        Args:
            callback_id (int): Callback Query ID
            data (str): Callback Data
            message (:class:`bale.Message`): Callback Message
            inline_message_id (str): Callback inline message id
            from_user (:class:`bale.User`): Callback The user who gave the callback.
            bot (:class:`bale.Bot`): Defaults to None.
    """
    __slots__ = (
        "message",
        "inline_message_id",
        "from_user",
        "data",
        "bot",
        "callback_id"
    )

    def __init__(self, callback_id: int = None, data: str = None, message: "Message" = None, inline_message_id: str = None,
                 from_user: "User" = None, bot: "Bot" = None):
        self.data = data
        self.callback_id = callback_id
        self.message = message
        self.inline_message_id = inline_message_id
        self.from_user = from_user
        self.bot = bot

    @classmethod
    def from_dict(cls, data: dict, bot: "Bot"):
        """
        Args:
            data (dict): Data
            bot (:class:`bale.Bot`): Bot
        """
        return cls(bot=bot, data=data.get("data"), callback_id=data.get("id"), message=Message.from_dict(data.get("message"), bot=bot),
                   inline_message_id=data.get("inline_message_id"), from_user=User.from_dict(bot=bot, data=data.get("from")))

    def to_dict(self):
        """Convert Class to dict
            Returns:
                :dict:
        """
        data = {
            "id": self.callback_id
        }

        if self.data:
            data["data"] = self.data
        if self.inline_message_id:
            data["inline_message_id"] = self.inline_message_id
        if self.message:
            data["message"] = self.message.to_dict()
        if self.from_user:
            data["from_user"] = self.from_user.to_dict()

        return data

    def __repr__(self):
        return f"<CallbackQuery inline_message_id={self.inline_message_id} message={self.message} from_user={self.from_user} data={self.data}>"
